/************************************************************************
 * 			MEKATRONIX Copyright 1998 		        *
 * Title        avoidtjp.c                                              *
 * Programmer	Keith L. Doty and modified for IC by Scott Jantz     	*
 * Date         August 25, 1998                                         *
 * Version	1			                		*
 *								        *
 * Description								*
 *    A very simple collision avoidance program. TJ PRO will read 	*
 *    each IR detectors, and turn away from any obstacles in its	*
 *    path. Also, if something hits TJ PRO's bumper, it will back up,	*
 *    turn, and go on.							*
 *			                                                *     					*
 ************************************************************************/
 

/*************************** Includes ********************************/

#include <tjpbase.h>
#include <stdio.h>

/************************ End of Includes ****************************/


/*************************** Constants ********************************/

#define AVOID_THRESHOLD 95

/************************ End of Constants ****************************/


/*************************** Prototypes *********************************/
void turn(void);
/************************ End of Prototypes *****************************/

void main(void)
/****************************** Main ***********************************/
{
  int irdr, irdl, speedr, speedl;

  init_analog();
  init_motortjp();
  init_clocktjp();

  IRE_ON;     /* turn on IR emitters */


 // START;  /*Press the rear bumper to start the program*/
  
   while(1)
  {

/*
   The following block will read the IR detectors, and decide whether TJ
   needs to turn to avoid any obstacles
*/
    irdr = RIGHT_IR;
    irdl = LEFT_IR;

    if (irdl > AVOID_THRESHOLD)
      speedr = -MAX_SPEED;
    else
      speedr = MAX_SPEED;
    if (irdr > AVOID_THRESHOLD)
      speedl = -MAX_SPEED;
    else
      speedl = MAX_SPEED;

    motorp(RIGHT_MOTOR, speedr); 
    motorp(LEFT_MOTOR, speedl);

/* This "if" statement checks the bumper. If the bumper is pressed, */
/* Tj will back up, and turn.                                       */

    if(FRONT_BUMP)
    {
      motorp(LEFT_MOTOR, -MAX_SPEED);
      motorp(RIGHT_MOTOR, -MAX_SPEED);
      wait(600);
      turn();
    }

    wait(35);
  }    
}
/**************************** End of Main ******************************/

void turn(void)
/************************************************************************
 * Function:  Will turn in a random direction for a "random" amount of  *
 *            time, dictated by the fast changine lower bits in		*
 *	      mseconds().  					        *
 * Returns:  None                                                       *
 *                                                                      *
 * Inputs                                                               *
 *   Parameters: None                                                   *
 *   Globals:    None                                                   *
 *   Registers:  TCNT                                                   *
 * Outputs                                                              *
 *   Parameters: None                                                   *
 *   Globals:    None                                                   *
 *   Registers:  None                                                   *
 * Functions called: motorp(), wait()					*
 * Notes:                                                               *
 ************************************************************************/
{
 int i;
 unsigned rand;

 rand = TCNT;

 if (rand & 0x0001)
 /*turn left*/
 {
   motorp(RIGHT_MOTOR, MAX_SPEED);
   motorp(LEFT_MOTOR, -MAX_SPEED);
 }
 else
 /*turn right*/
 {
   motorp(RIGHT_MOTOR, -MAX_SPEED);
   motorp(LEFT_MOTOR, MAX_SPEED);
 }

 i=(rand % 1024);
 if(i>250) wait(i); else wait(250);

}

/***********************End Function turn ****************************/


