/************************************************************************
* LINESDG by David Grindlinger
* Copyright 2000
*
* a basic line finding and following code for my Mekatronix TJPro-based
* robot. Code adapted with functions reproduced from software AVOIDTJP by 
* Keith Doty, with Scott Jantz, 1998. 
 ************************************************************************/
 

  //includes
	#include <tjpbasdg.h>
	#include <stdio.h>

  //constants
	#define TIMEOUT 200 // how many iterations to carry
						// out an activity before timing out
						
	#define OBSTACLE 105 // avoidance threshhold for IR detectors
	
  //prototypes
	void adjustLeft(void);
	void adjustRight(void);
	void turn(void);
	void locate(void);
	void follow(void);
	void calibrate(void);
	void pollcells(void);
	void pens_up(void);
	void left_pen_down(void);
	void right_pen_down(void);
	void getcolorvalue (void);
	void behavior(void);
	void redtest(void);
	void greentest(void);
	void avoid(void);
	void followtrail(void);
	void dropfood(void);
	void maketrail(void);
	void aboutface(void);

	
  //global variables
	int WHITE_LT, WHITE_RT, WHITE_MD; //line sensor threshholds
	int WHITE_BL, WHITE_GR, WHITE_RD;
	int GOTHIVE=0;  //flag true when trail made/found from hive to food
	int GOTFOOD =0; // //flags true when going from food to hive
	int FOLLOWING=0;	//flag is true when following a line
	int EVENT=0;	//true when changing between lines of diff. colors
	//int speedl, speedr; //current motor speed
	int cds_lt,cds_rt,cds_md,cds_rd,cds_gr,cds_bl;//latest cds sensor values
	int color; //color under color detector, returned by getcolor()
	int irdl, irdm,irdr;//latest IR detector values	
	
	
//MAIN
	void main(void)
	{

 	 //init library functions
		init_analog();
  	  //init_motortjp();
		init_servotjp();
		init_clocktjp();
	  //put pens in neutral
  	  	pens_up();
  	  //turn off LEDs
  	  	y4=led_state;
  	  //put bot in motion
  		servo(rightmotor,rightspeed);
  		servo(leftmotor,leftspeed);
     		wait(400);
  	  //calibrate sensors over white area of floor
    		calibrate();
  	
  	
	
  while(1)
  {
    //search loop--look for lines 
  	while(!FOLLOWING)
  	{
	  //avoid collisions
		avoid();
	  
	  //poll cds cells
	  //if line detected, remember who saw it, trigger locate behavior	
		pollcells();		
		if (LINE) locate();
	}

	  //robot probably on a line to have returned from following routine
	  //should be positioned with color detector on line . whats is it?
		if (color=BLACK) followtrail();
		if (color=RED) hivebehavior();
		if (color=GREEN) foodbehavior();

	  //just in case-- reset search loop;
		FOLLOWING=0;
		
    wait(35); //here to avoid signalling motors too frequently
  }

		


}//end function MAIN


//LOCATE
void locate(void)
{
  //variables
	int hits=1; //sensor hit counter
	int c=0; //loop counter

  	//wait for the last sensor to see the line
  	//turn robot towards that sensor to follow line
  	//if wait too long, forget it, return to search
  	while ((c<TIMEOUT) && (hits<3))
  	{
  		avoid();
		pollcells();
		if (LINE) hits++;
		c++;
		wait(25);
	}		
	
	//third hit means time to aquire line
	//if c=TIMEOUT, time-out the search
	if (c>TIMEOUT) {}
	
	//otherwise turn away from last sensor to hit line
	else
	{	
		if(WHITE_LT - cds_lt < WHITE_RT - cds_rt) adjustRight();
		else adjustLeft();
		
	  //when the program returns from adjustleft or right,
	  //robot should have center cds_cell over target and should be
	  //going straight	
		pollcells();
		getcolorvalue();
	}
}//end LOCATE



//FOLLOW
void follow(void)
{
	//remember current line color
	int linecolor=color;	

	//go forward
    servo(leftmotor,leftspeed);
    servo(rightmotor,rightspeed);
	
	while (FOLLOWING)
	{
		//when stray, make corrections
		pollcells();
		if (cds_lt<WHITE_LT) adjustLeft();
		if (cds_rt<WHITE_RT) adjustRight();
		wait(30);
		//if hit line of different color
		//leave follow behavior
		if (COLOREDLINE) 
		{
			getcolorvalue();
			if (color!=linecolor) 
			{
				FOLLOWING=0;
				EVENT=1;
			}
		}
	}
}//end FOLLOW


//ADJUSTLEFT
void adjustLeft(void)
{
	int c=0; //loop count

  //stop or reverse one motor until the some cds cell
  //  passes over the line, or times out
	servo(leftmotor,0);
	
	while (c<TIMEOUT)
	{
		pollcells();
        
        //if center cds sensor crosses line, go straight again
        if (cds_md<WHITE_MD)
        {
        
        	servo(leftmotor,leftspeed);
        	break;
        }	
        
        //if left cds sensor hits, adjust right
        else if (cds_lt<WHITE_LT)
        {
        
            servo(leftmotor,leftspeed);
        	adjustRight();
        	break;
        }	
        
        //if right cds cell hit, adjust left
       else if (cds_rt<WHITE_RT)
        {
            servo(leftmotor,leftspeed);
        	adjustLeft();
        	break;
        }	        	

		wait(20);
		c++;
	}	
	//if timed out, turn off following flag
	if (c>=TIMEOUT) FOLLOWING=0;
    servo(leftmotor,leftspeed);
}//end ADJUSTLEFT


//ADJUSTRIGHT
void adjustRight(void)
{
	int c=0; //loop count

  //stop or reverse one motor until the some cds cell
  //  passes over the line, or times out
	servo(rightmotor,0); 
	
	while(c<TIMEOUT)
	{
		pollcells();
        
        //if center cds sensor crosses line
        //determine color of line
        //then go straight again
        if (cds_md<WHITE_MD)
        {
        	getcolorvalue();
        	servo(rightmotor,rightspeed);
        	break;
        	//test lines below
        }	
        
        //if left cds sensor hits, adjust right
        if (cds_lt<WHITE_LT)
        {
          	servo(rightmotor,rightspeed);
        	adjustRight();
        	break;
        	//test lines below
        }	
        
        //if right cds cell hit, adjust left
        if (cds_rt<WHITE_RT)
        {
            servo(rightmotor,rightspeed);
        	adjustLeft();
        	break;
        	//test lines below
        }	        	

		wait(20);
		c++;
	}	
	//if timed out, turn off following flag
	
    servo(rightmotor,rightspeed);
	if (c>=TIMEOUT) FOLLOWING=0;
	
}//end ADJUSTRIGHT






void maketrail(void)
{
	while (FOLLOWING)
	{
		pollcells();
		if (cds_lt<WHITE_LT) adjustLeft();
		if (cds_rt<WHITE_RT) adjustRight();
		if ((!LINE)&&(!COLOREDLINE)) break;
		wait(30);
	}
	servo(leftmotor,0);
	servo(rightmotor,0);
	left_pen_down();
	wait(500);
	servo(leftmotor,leftspeed);
	servo(rightmotor,rightspeed);
	while (!GOTHIVE)
	{
		avoid();
		pollcells();
		if (COLOREDLINE) 
		{
			getcolorvalue();
			if (color==RED)
			{
				pens_up();
				dropfood();
				GOTHIVE=1;
			}
			if (color==BLACK) followtrail();
			if (color==GREEN) behavior();	
		}
	}
}//end MAKETRAIL



//FOLLOWTRAIL
void followtrail(void)
{
   //follow current line
	FOLLOWING=1;
	follow();
  //when come back, either lost or hit new color
			
}




//DROP FOOD
void dropfood (void)
{
	GOTFOOD=0;
	aboutface();
}
//END DROP FOOD



//ABOUTFACE
void aboutface(void)
{
	//start turning around, pass line with cds cells
	servo(leftmotor,leftreverse);
	wait(300);
	//keep turning until see line behind you
	pollcells();
	while (!line) 
	{
		wait(20); 
		pollcells();
	}
	//use locate command to center on line
	locate;	
} 
//END ABOUT FACE



//PEN CONTROL FUNCTIONS
//operate servo controlling pens
void left_pen_down(void)
{ 
	servo(2,3100);
}

void right_pen_down(void)
{  
   	 servo(2,1900); 
}
   
void pens_up(void)
{
   	servo(2,2500);    
}//end PEN CONTROL FUNCTIONS



//POLLCELLS
void pollcells(void)
//refresh values of cds cell readings
//MUX = analog input multiplexor/expandor board
{		
	LEFT_CDS;
	cds_lt=MUX;
	RIGHT_CDS;
	cds_rt=MUX;
	GREEN_CDS;
	cds_md=cds_gr=MUX;
	BLUE_CDS;
	cds_bl=MUX;
	RED_CDS;
	cds_rd=MUX;
}//end of POLLCELLS



//GETCOLORVALUE
//determine color with red/green filtered cds cells
void getcolorvalue(void)
{
	int rdistance, rdist_sq, gdistance, gdist_sq,colorvalue;

  //distances = distance of filtered cds_cell from white

	rdistance=WHITE_RD-cds_rd;
	gdistance=WHITE_GR-cds_gr;

  //normalize distances
  	rdistance=rdistance*120/WHITE_RD;
  	gdistance=gdistance*120/WHITE_GR;
	
	colorvalue=rdistance+gdistance;
	
	if (colorvalue<40) color=WHITE;
	else if (colorvalue>130) color = BLACK;
	else if (gdistance>rdistance) color = RED;
	else color=GREEN;
	
}//end GETCOLORVALUE
	


//AVOID
void avoid(void)
//checks IR for obstacles
//and bumpers for collisions
{

  //check for collision and
  //look with IR for upcoming collisions
  //if bumped, or center IR detects object, backup and turn
  	
	if ( (MIDDLE_IR>OBSTACLE) || (_BUMP) )
	{
		FOLLOWING=0;
   		servo(leftmotor,leftreverse);
    		servo(rightmotor,rightreverse);
    		wait(500);
		turn();
    		servo(leftmotor,leftspeed);
    		servo(rightmotor,rightspeed);   
    	}
    

  //if detect object to side, turn away
	while (LEFT_IR>OBSTACLE) 
	{
		FOLLOWING=0;
		servo(rightmotor, rightreverse);
		wait(200);
	}
	
	while (RIGHT_IR>OBSTACLE)
	{
		FOLLOWING=0;
		servo(leftmotor, leftreverse);
		wait(200);
	}
	
}// END AVOID



//TURN
void turn(void)
{
	int i;
 	unsigned rand;

 	rand = TCNT;
 	if (rand & 0x0001)
 	/*turn left*/
 	{
  		servo(rightmotor, rightspeed);
  	 	servo(leftmotor, leftreverse);
 	}
 	else
 	/*turn right*/
 	{
   		servo(rightmotor, rightreverse);
   		servo(leftmotor, leftspeed);
 	}

 	i=(rand % 1024);
 	if(i>250) wait(i); else wait(250);
}//end of TURN



//CALIBRATE
void calibrate(void)
  //establish threshold for line/color detection on surface
{  
  //calibrate line detection threshhold for cds cells
	int i=0; //loop counter
	WHITE_LT=WHITE_RT=WHITE_MD=0; //initialize sensors values
	WHITE_RD=WHITE_GR=WHITE_BL=0;
	
  //for each cds cell: take 3 readings 	
  //with a delay between each.
	while(i<3)
	{
		pollcells();
		WHITE_LT=WHITE_LT+cds_lt;
		WHITE_RT=WHITE_RT+cds_rt;
		WHITE_MD=WHITE_MD+cds_md;
		WHITE_BL=WHITE_BL+cds_bl;
		WHITE_RD=WHITE_RD+cds_rd;
		wait(25);
		i++;
	}
	      
  //Find mean.  
  //Knock off 10 points as a noise buffer,
  //set as threshhold for line detection 
	WHITE_LT=WHITE_LT/3 - 10;
	WHITE_RT=WHITE_RT/3 - 10;
	WHITE_GR=WHITE_MD=WHITE_MD/3 - 10;
	WHITE_BL=WHITE_BL/3 - 10;
	WHITE_RD=WHITE_RD/3 - 10;

}//end of CALIBRATE
	


