/************************************************************************
* LINESDG by David Grindlinger
* Copyright 2000
*
* a basic line finding and following code for my Mekatronix TJPro-based
* robot. Code adapted with functions reproduced from software AVOIDTJP by a
* Keith Doty, with Scott Jantz, 1998. 
 ************************************************************************/
 

  //includes
	#include <tjpbasdg.h>
	#include <stdio.h>

  //constants
	#define TIMEOUT 175 // how many iterations to carry
						// out an activity before timing out
						
	#define OBSTACLE 100 // avoidance threshhold for IR detectors
	
  //prototypes
	void adjustLeft(void);
	void adjustRight(void);
	void turn(void);
	void locate(void);
	void follow(void);
	void calibrate(void);
	void pollcells(void);
	void pens_up(void);
	void left_pen_down(void);
	void right_pen_down(void);
	void getcolorvalue (void);
	void hivebehavior(void);
	void foodbehavior(void);
	void avoid(void);
	void followtrail(void);
	void maketrail(void);
	void aboutface(void);
	void do180(void);

	
  //global variables
	int WHITE_LT, WHITE_RT, WHITE_MD; //line sensor threshholds
	int WHITE_BL, WHITE_GR, WHITE_RD;
	int GOTHIVE=0;  //flag true when trail made/found from hive to food
	int EVENT=0; //true if changing from one lien to another
	int GOTFOOD =0; // //flags true when going from food to hive
	int FOLLOWING=0;	//flag is true when following a line
	int cds_lt,cds_rt,cds_md,cds_rd,cds_gr,cds_bl; //latest cds sensor values
	int color=WHITE; //color under color detector, returned by getcolor()

	

//MAIN
	void main(void)
	{

 	 //init library functions
		init_analog();
		init_servotjp();
		init_clocktjp();
	  //put pens in neutral
  	  	pens_up();
		wait(100);
  	  //put bot in motion
  		servo(rightmotor,rightspeed);
  		servo(leftmotor,leftspeed);
     	wait(400);
  	  //calibrate sensors over white area of floor
    	calibrate();
  	

  //search loop--look for lines 
  while(1)
  {
  	printf("main\n");
  	while(!FOLLOWING)
  	{
	  //avoid collisions
		avoid();
	  //poll cds cells
	  //if line detected, aquire with locate behavior
		pollcells();
		if (LINE) printf("m:line\n");
		if (LINE) locate();
	}
	  //robot probably on a line to have returned from locate routine
	  //should be positioned with color detector on line . whats is it?
	  	printf("m:color = %d\n",color);
		if (color==BLACK) {printf("m:bk-f_trl\n");followtrail();}
		if (color==RED) {printf("m:rd-hv_bhvr\n");hivebehavior();}
		if (color==GREEN) {printf("m:gr-fd_bhvr\n");foodbehavior();}

	  //if not on line-- restart search loop

    wait(35); //here to avoid signalling motors too frequently
  }
}//end function MAIN



//LOCATE
void locate(void)
{

  //variables
	int hits=1; //sensor hit counter
	int c=0; //loop counter
	printf("locate\n");
  	//wait for the last sensor to see the line
  	//turn robot towards that sensor to follow line
  	//if wait too long, forget it, return to search
  	while ((c<TIMEOUT) && (hits<3))
  	{
  		avoid();
		pollcells();
		if (LINE) hits++;
		c++;
		//get color if possible
		if (COLOREDLINE) 
		{
			printf("lct: ");
			getcolorvalue();
			//if color =red or green, escape locate routine
			if ((color==RED)||(color==GREEN)) 
			{
				FOLLOWING=1;
				c=TIMEOUT+1;
			}
		}
	}		
	
	//third hit means time to aquire line
	//if green or red, return to main to decide on behavior
	//if c=TIMEOUT, time-out the search
	if (c>TIMEOUT){}

	
	//otherwise turn away from last sensor to hit line
	else
	{	
		printf("lct_aquire: ");
		if(WHITE_LT - cds_lt < WHITE_RT - cds_rt) adjustRight();
		else adjustLeft();
		
	  //when the program returns from adjustleft or right,
	  //robot should have center cds_cell over target and should be
	  //going straight	
		pollcells();
		getcolorvalue();
		FOLLOWING=1;
	}
}//end LOCATE



//FOLLOW
void follow(void)
{

	//remember current line color
	int linecolor=color;	

	//go forward
    servo(leftmotor,leftspeed);
    servo(rightmotor,rightspeed);
	
	while (FOLLOWING)
	{
		printf("following line %d\n",linecolor);
		//when stray, make corrections
		pollcells();
		if (cds_lt<WHITE_LT) {printf("adj-L\n");adjustLeft();}
		if (cds_rt<WHITE_RT) {printf("adj-R\n");adjustRight();}
		wait(25);
		//if hit line of different color
		//leave follow behavior
		printf("flwng: ");getcolorvalue();
		if (color!=linecolor) FOLLOWING=0;
	}
	printf("following-0,color=%d\n",color);
}//end FOLLOW


//ADJUSTLEFT
void adjustLeft(void)
{
	int c=0; //loop count

  //stop or reverse one motor until the some cds cell
  //  passes over the line, or times out
	servo(leftmotor,0);
	printf("adj-L\n");
	while (c<TIMEOUT)
	{
		pollcells();
        
        //if center cds sensor crosses line, go straight again
		if (LINE) servo(leftmotor,leftspeed);
        if (cds_md<WHITE_MD)
        {
        	break;
        }	
        
        //if left cds sensor hits, adjust right
        else if (cds_lt<WHITE_LT)
        {
        	adjustRight();
        	break;
        }	
        
        //if right cds cell hit, adjust left
       else if (cds_rt<WHITE_RT)
        {
        	adjustLeft();
        	break;
        }	        	

		wait(20);
		c++;
	}	

    servo(leftmotor,leftspeed);

	//if timed out, turn off following flag
	if (c>=TIMEOUT) 
	{
		printf("timeout adj-L\n");
		FOLLOWING=0;
		color=WHITE;
	}

}//end ADJUSTLEFT


//ADJUSTRIGHT
void adjustRight(void)
{
	int c=0; //loop count

  //stop or reverse one motor until the some cds cell
  //  passes over the line, or routine times out
	servo(rightmotor,0); 
	printf("adj-R\n");
	while(c<TIMEOUT)
	{
		pollcells();
        
        //if center cds sensor crosses line
        //then go straight again
		if (LINE) servo(rightmotor,rightspeed);
        if (cds_md<WHITE_MD)
        {
        	break;
        }	
        
        //if left cds sensor hits, adjust right
        if (cds_lt<WHITE_LT)
        {
        	adjustRight();
        	break;
        }	
        
        //if right cds cell hit, adjust left
        if (cds_rt<WHITE_RT)
        {
        	adjustLeft();
        	break;
        }	        	

		wait(20);
		c++;
	}	
	//if timed out, turn off following flag
	
    servo(rightmotor,rightspeed);
	if (c>=TIMEOUT) 
	{
		printf("timeout adj-R\n");
		FOLLOWING=0;
		color=WHITE;
	}
	
}//end ADJUSTRIGHT



//FOLLOWTRAIL
void followtrail(void)
{
	printf("followtrail\n");	
   //follow current line
	follow();
  //when come back, either lost trail or hit new color
  //if found hive or food, change to appropraite behavior
	if (color==RED) hivebehavior();
	else if (color==GREEN) foodbehavior();
  //if lost line, go back to main loop
}



//HIVEBEHAVIOR
void hivebehavior(void)
{
	printf("hivebehavior\n");
	//if you are drawing, pens up
	//if you have food, drop it, go back to food
	if (GOTFOOD)
	{
		wait(100);
		pens_up();
		GOTHIVE=1;
		GOTFOOD=0;
		aboutface();
		if (FOLLOWING) followtrail();
	}
	
	//if no food, turn 'em out to go find some
	else 
	{
		FOLLOWING=0;
   		servo(leftmotor,leftreverse);
    		servo(rightmotor,rightreverse);
    		wait(500);
		turn();
    		servo(leftmotor,leftspeed);
    		servo(rightmotor,rightspeed);   
    	}
}//end HIVEBEHAVIOR



//FOODBEHAVIOR
void foodbehavior(void)
{
	printf("foodbehavior\n");
	//get some food
	GOTFOOD=1;

	//if coming from hive, turnaround
	if (GOTHIVE)
	{
		aboutface();
		if (FOLLOWING) followtrail();
		//retrace line as returning
		left_pen_down;
	}

	//if no food, follow green "foodline" until end
	//start drawing line back to food

	else
	{
		maketrail();
	}
}



//MAKETRAIL
void maketrail(void)
//draw the trail from food to hive
{
	int DRAWING=1;
	printf("maketrail\n");
	left_pen_down();
	wait(125);
	do180();

	
	while(DRAWING)
	{
		pollcells();
		if(COLOREDLINE)
		{
			getcolorvalue();
			if (color!=WHITE)
			{
				DRAWING=0;
				pens_up();
			}
		}
		wait(20);
	}
}
//end MAKETRAIL



//ABOUTFACE
void aboutface(void)
{
	int c=0;
	printf("aboutface\n");
	
	//start turning around, pass line with cds cells
	servo(leftmotor,leftreverse);
	wait(100);
	//keep turning until see line behind you
	pollcells();
	while (!COLOREDLINE)
	{
		wait(20); 
		pollcells();
		c++;
	}
	//use locate command to center on line
	FOLLOWING=1;
} 
//END ABOUT FACE



//PEN CONTROL FUNCTIONS
//operate servo controlling pens
void left_pen_down(void)
{ 
	servo(2,3100);
   	printf("lt-pens\n");
}

void right_pen_down(void)
{  
   	 servo(2,1900); 
   	printf("rt_pen\n");
}
   
void pens_up(void)
{
   	servo(2,2500);    
   	printf("pens_up\n");
}//end PEN CONTROL FUNCTIONS



//POLLCELLS
void pollcells(void)
//refresh values of cds cell readings
//MUX = analog input multiplexor/expandor board
{		
	LEFT_CDS;
	cds_lt=MUX;
	RIGHT_CDS;
	cds_rt=MUX;
	GREEN_CDS;
	cds_md=cds_gr=MUX;
	BLUE_CDS;
	cds_bl=MUX;
	RED_CDS;
	cds_rd=MUX;
}//end of POLLCELLS



//GETCOLORVALUE
//determine color with red/green filtered cds cells
void getcolorvalue(void)
{
	int rdistance, rdist_sq, gdistance, gdist_sq,colorvalue;
	printf("color = ");
  //distances = distance of filtered cds_cell from white

	rdistance=WHITE_RD-cds_rd;
	gdistance=WHITE_GR-cds_gr;

  //normalize distances
  	rdistance=rdistance*120/WHITE_RD;
  	gdistance=gdistance*120/WHITE_GR;
	
	colorvalue=rdistance+gdistance;
	
	if (colorvalue<40) color=WHITE;
	else if (colorvalue>130) color = BLACK;
	else if (gdistance>rdistance) color = RED;
	else color=GREEN;
	printf("%d:  cv:%d  g-:d%  r-d:%",color,colorvalue,gdistance,rdistance); 
	
}//end GETCOLORVALUE
	


//AVOID
void avoid(void)
//checks IR for obstacles
//and bumpers for collisions
{

  //check for collision and
  //look with IR for upcoming collisions
  //if bumped, or center IR detects object, backup and turn
	if ( (MIDDLE_IR>OBSTACLE) || (_BUMP) )
	{
		FOLLOWING=0;
   		servo(leftmotor,leftreverse);
    		servo(rightmotor,rightreverse);
    		wait(500);
		turn();
    		servo(leftmotor,leftspeed);
    		servo(rightmotor,rightspeed);   
    	}
    

  //if detect object to side, turn away
	while (LEFT_IR>OBSTACLE) 
	{
		FOLLOWING=0;
		servo(rightmotor, rightreverse);
		wait(200);
	}
	
	while (RIGHT_IR>OBSTACLE)
	{
		FOLLOWING=0;
		servo(leftmotor, leftreverse);
		wait(200);
	}
	
}// END AVOID



//TURN
void turn(void)
{
	int i;
 	unsigned rand;

 	rand = TCNT;
 	if (rand & 0x0001)
 	/*turn left*/
 	{
  		servo(rightmotor, rightspeed);
  	 	servo(leftmotor, leftreverse);
 	}
 	else
 	/*turn right*/
 	{
   		servo(rightmotor, rightreverse);
   		servo(leftmotor, leftspeed);
 	}

 	i=(rand % 1024);
 	if(i>250) wait(i); else wait(250);
}//end of TURN



//CALIBRATE
void calibrate(void)
  //establish threshold for line/color detection on surface
{  
  //calibrate line detection threshhold for cds cells
	int i=0; //loop counter
	WHITE_LT=WHITE_RT=WHITE_MD=0; //initialize sensors values
	WHITE_RD=WHITE_GR=WHITE_BL=0;
	printf("calibrate\n");
	
  //for each cds cell: take 3 readings 	
  //with a delay between each.
	while(i<3)
	{
		pollcells();
		WHITE_LT=WHITE_LT+cds_lt;
		WHITE_RT=WHITE_RT+cds_rt;
		WHITE_MD=WHITE_MD+cds_md;
		WHITE_BL=WHITE_BL+cds_bl;
		WHITE_RD=WHITE_RD+cds_rd;
		wait(150);
		i++;
	}
	      
  //Find mean.  
  //Knock off 10 points as a noise buffer,
  //set as threshhold for line detection 
	WHITE_LT=WHITE_LT/3 - 10;
	WHITE_RT=WHITE_RT/3 - 10;
	WHITE_GR=WHITE_MD=WHITE_MD/3 - 10;
	WHITE_BL=WHITE_BL/3 - 10;
	WHITE_RD=WHITE_RD/3 - 10;

}//end of CALIBRATE
	
	
void do180(void)
//turn 180 degrees,halt
{
	printf("do180\n");
	servo(leftmotor,leftspeed);
	servo(rightmotor,rightreverse);
	wait(1250);
	servo(rightmotor,rightspeed);
	servo(leftmotor,leftspeed);
}

//notes: avoidance during line following, and w/food vs. w/out food
