/************************************************************************
* LINESDG by David Grindlinger
* Copyright 2000
*
* a basic line finding and following code for my Mekatronix TJPro-based
* robot. Code adapted with functions reproduced from software AVOIDTJP by 
* Keith Doty, with Scott Jantz, 1998. 
 ************************************************************************/
 

  //includes
	#include <tjpbasdg.h>
	#include <stdio.h>
	
  //constants
	#define TIMEOUT 500 // how many iterations to carry
						// out an activity before timing out
  //prototypes
	void adjustLeft(void);
	void adjustRight(void);
	void turn(void);
	void locate(void);
	void follow(void);
	void calibrate(void);
	void pollcells(void);
	void pens_up(void);
	void left_pen_down(void);
	void right_pen_down(void);
	void getcolor (void);
	
	
  //global variables
	int WHITE_LT, WHITE_RT, WHITE_MD; //line sensor threshholds
	int WHITE_BL, WHITE_GR, WHITE_RD;
	int FOLLOWING=0;	//flag is true when following a line
	int LASTSEEN;		//which cds cell saw a line last
					//LEFTSIDE,MIDDLE,RIGHTSIDE = left/right/middle
	int speedl, speedr; //current motor speed
	int cds_lt,cds_rt,cds_md,cds_rd,cds_gr,cds_bl;//latest cds sensor values
	int color; //color under color detector, returned by getcolor()
//MAIN
void main(void)
{

  //init library functions
	init_analog();
  //init_motortjp();
	init_servotjp();
	init_clocktjp();
	
	
  //set pens at neutral
  	left_pen_down();
  	wait(500);
    pens_up();
    wait(500);	
  //put bot in motion
    servo(rightmotor,rightspeed);
    servo(leftmotor,leftspeed);
    wait(400);
    
  //calibrate sensors over white area of floor
  //establish threshold for line detection on white
  	calibrate();


   //search loop--look for lines 	
  	while(1)
  	{

	  //poll cds cells
	  //if line detected, remember who saw it, trigger locate behavior	
		pollcells();		
		if (LINE)
			locate();

		
		
	  /* This "if" statement checks the bumper. If the bumper is pressed, */
	  /* Tj will back up, and turn.                                       
   		if(FRONT_BUMP)
    	{
    		servo(leftmotor,leftreverse);
        	servo(rightmotor,rightreverse);
        	wait(800);
        	turn();
    		servo(leftmotor,leftspeed);
        	servo(rightmotor,rightspeed);        	
   		}*/
		
    wait(35); //here to avoid signalling motors to frequently
    
  	}    
  	
}//end function MAIN


//LOCATE
void locate(void)
{
  //variables
	int hits=1; //sensor hit counter
	int c=0; //loop counter

  	//wait for the last sensor to see the line
  	//turn robot towards that sensor to follow line
  	//if wait too long, forget it, return to search
  	while ((c<TIMEOUT) && (hits<3))
  	{
		pollcells();
		if (LINE)
			hits++;
		c++;
		wait(25);
	}		
	
	//third hit means time to aquire line
	//if c=TIMEOUT, time-out the search
	if (c>TIMEOUT) {}
	
	//otherwise turn away from last sensor to hit line
	else
	{	
		if(WHITE_LT - cds_lt < WHITE_RT - cds_rt) adjustRight();
		else adjustLeft();
		FOLLOWING=1;
		follow();
		
	}
}//end LOCATE


//FOLLOW
void follow(void)
{
	int stray =0; //flag for when robot strays from line
	
    servo(leftmotor,leftspeed);
    servo(rightmotor,rightspeed);
	
	while (FOLLOWING==1)
	{
		pollcells();
		if (cds_lt<WHITE_LT) adjustLeft();
		if (cds_rt<WHITE_RT) adjustRight();
		wait(30);
	}
	
}//end FOLLOW


//ADJUSTLEFT
void adjustLeft(void)
{
	int c=0; //loop count

  //stop or reverse one motor until the some cds cell
  //  passes over the line, or times out
	servo(leftmotor,0);
	
	while(c<TIMEOUT)
	{
		pollcells();
        
        //if center cds sensor crosses line, go straight again
        if (cds_md<WHITE_MD)
        {
        
        	servo(leftmotor,leftspeed);
        	break;
        }	
        
        //if left cds sensor hits, adjust right
        else if (cds_lt<WHITE_LT)
        {
        
            servo(leftmotor,leftspeed);
        	adjustRight();
        	break;
        }	
        
        //if right cds cell hit, adjust left
       else if (cds_rt<WHITE_RT)
        {
            servo(leftmotor,leftspeed);
        	adjustLeft();
        	break;
        }	        	

		wait(20);
		c++;
	}	
	//if timed out, turn off following flag
	if (c>=TIMEOUT) FOLLOWING=0;
    servo(leftmotor,leftspeed);
}//end ADJUSTLEFT


//ADJUSTRIGHT
void adjustRight(void)
{
	int c=0; //loop count

  //stop or reverse one motor until the some cds cell
  //  passes over the line, or times out
	servo(rightmotor,0); 
	
	while(c<TIMEOUT)
	{
		pollcells();
        
        //if center cds sensor crosses line, go straight again
        if (cds_md<WHITE_MD)
        {
        	servo(rightmotor,rightspeed);
        	break;
        	//test lines below
        }	
        
        //if left cds sensor hits, adjust right
        if (cds_lt<WHITE_LT)
        {
          	servo(rightmotor,rightspeed);
        	adjustRight();
        	break;
        	//test lines below
        }	
        
        //if right cds cell hit, adjust left
        if (cds_rt<WHITE_RT)
        {
            servo(rightmotor,rightspeed);
        	adjustLeft();
        	break;
        	//test lines below
        }	        	

		wait(20);
		c++;
	}	
	//if timed out, turn off following flag
	
    servo(rightmotor,rightspeed);
	if (c>=TIMEOUT) FOLLOWING=0;
	
}//end ADJUSTRIGHT


//CALIBRATE
void calibrate(void)
{  
  //calibrate line detection threshhold for cds cells
	int i=0; //loop counter
	WHITE_LT=WHITE_RT=WHITE_MD=0; //initialize sensors values
	WHITE_RD=WHITE_GR=WHITE_BL=0;
	
  //for each cds cell: take 3 readings 	
  //with a delay between each.
	while(i<3)
	{
		pollcells();
		WHITE_LT=WHITE_LT+cds_lt;
		WHITE_RT=WHITE_RT+cds_rt;
		WHITE_MD=WHITE_MD+cds_md;
		WHITE_BL=WHITE_BL+cds_bl;
		WHITE_RD=WHITE_RD+cds_rd;
		wait(25);
		i++;
	}
	      
  //Find mean.  
  //Knock off 10 points as a noise buffer,
  //set as threshhold for line detection 
	WHITE_LT=WHITE_LT/3 - 10;
	WHITE_RT=WHITE_RT/3 - 10;
	WHITE_GR=WHITE_MD=WHITE_MD/3 - 10;
	WHITE_BL=WHITE_BL/3 - 10;
	WHITE_RD=WHITE_RD/3 - 10;

}//end of CALIBRATE
	

//TURN
void turn(void){
 int i;
 unsigned rand;

 rand = TCNT;

 if (rand & 0x0001)
 /*turn left*/
 {
   servo(rightmotor, rightspeed);
   servo(leftmotor, leftreverse);
 }
 else
 /*turn right*/
 {
   	servo(rightmotor, rightreverse);
   	servo(leftmotor, leftspeed);
 }

 i=(rand % 1024);
 if(i>250) wait(i); else wait(250);

}//end of TURN


//PEN CONTROL FUNCTIONS
void left_pen_down(void)
{ 
	servo(2,3100);
}

void right_pen_down(void)
{  
   	 servo(2,1900); 
}
   
void pens_up(void)
{
   	servo(2,2500);    
}

//POLLCELLS
void pollcells(void)
{		
	LEFT_CDS;
	cds_lt=MUX;
	RIGHT_CDS;
	cds_rt=MUX;
	GREEN_CDS;
	cds_md=cds_gr=MUX;
	BLUE_CDS;
	cds_bl=MUX;
	RED_CDS;
	cds_rd=MUX;
}//end of POLLCELLS


//GETCOLOR
void getcolor(void)
{
	int rd_neg, gr_neg;
	
	rd_neg=WHITE_RD-cds_rd;
	gr_neg=WHITE_GR-cds_gr;
}


	
	
	