/************************************************************************
* LINESDG by David Grindlinger
* Copyright 2000
*
* a basic line finding and following code for my Mekatronix TJPro-based
* robot. Code adapted with functions reproduced from software AVOIDTJP by 
* Keith Doty, with Scott Jantz, 1998. 
 ************************************************************************/
 

  //includes
	#include <tjpbasdg.h>
	#include <stdio.h>
	
  //constants
	#define TIMEOUT 500 // how many iterations to carry
						// out an activity before timing out
  //prototypes
	void adjustLeft(void);
	void adjustRight(void);
	void turn(void);
	void locate(void);
	void follow(void);
	void calibrate(void);

  //global variables
	int WHITE_LT, WHITE_RT, WHITE_MD; //line sensor threshholds
	int FOLLOWING=0;	//flag is true when following a line
	int LASTSEEN;		//which cds cell saw a line last
					//LEFTSIDE,MIDDLE,RIGHTSIDE = left/right/middle
	int speedl, speedr; //current motor speed

	
//MAIN
void main(void)
{
  //variables
	int cds_lt, cds_rt, cds_md;  //latest cds sensor values

  //init library functions
	init_analog();
	init_motortjp();
	init_clocktjp();
	
  //put bot in motion
	speedr=SLOW_SPEED; speedl=speedr;
    motorp(RIGHT_MOTOR, speedr+10); 
    motorp(LEFT_MOTOR, speedl );
    wait(400);
    
  //calibrate sensors over white area of floor
  //establish threshold for line detection on white
  	calibrate();


   //search loop--look for lines 	
  	while(1)
  	{

	  //poll cds cells
	  //if line detected, remember who saw it, trigger locate behavior	
		LEFT_CDS;
		cds_lt=MUX;
		MIDDLE_CDS;
		cds_md=MUX;
		RIGHT_CDS;
		cds_rt=MUX;			
		if ((cds_lt<WHITE_LT)||(cds_rt<WHITE_RT)||(cds_md<WHITE_MD))
			locate();

		
		
	  /* This "if" statement checks the bumper. If the bumper is pressed, */
	  /* Tj will back up, and turn.                                       */
    	if(FRONT_BUMP)
    	{
    		motorp(LEFT_MOTOR, speedl=-MAX_SPEED);
        	motorp(RIGHT_MOTOR, speedr=-MAX_SPEED);
        	wait(800);
        	turn();
    		motorp(LEFT_MOTOR, speedl=SLOW_SPEED );
        	motorp(RIGHT_MOTOR, speedr=SLOW_SPEED+10);        	
   		}

    wait(35); //here to avoid signalling motors to frequently
    
  	}    
  	
}//end function MAIN


//LOCATE
void locate(void)
{
  //variables
	int cds_lt, cds_rt, cds_md;  //latest cds sensor values
	int hits=1; //sensor hit counter
	int c=0; //loop counter

  	//wait for the last sensor to see the line
  	//turn robot towards that sensor to follow line
  	//if wait too long, forget it, return to search
  	while ((c<TIMEOUT) && (hits<3))
  	{
		LEFT_CDS;
		cds_lt=MUX;
		MIDDLE_CDS;
		cds_md=MUX;
		RIGHT_CDS;
		cds_rt=MUX;		
		if ((cds_lt<WHITE_LT)||(cds_rt<WHITE_RT)||(cds_md<WHITE_MD))
			hits++;
		c++;
		wait(25);
	}		
	
	//third hit means time to aquire line
	//if c=TIMEOUT, time-out the search
	if (c>TIMEOUT) {}
	
	//otherwise turn away from last sensor to hit line
	else
	{	
		if(WHITE_LT - cds_lt < WHITE_RT - cds_rt) adjustRight();
		else adjustLeft();
		FOLLOWING=1;
		follow();
		
	}
}//end LOCATE


//FOLLOW
void follow(void)
{
	int stray =0; //flag for when robot strays from line
	int cds_lt, cds_rt, cds_md; //latest cds sensor values
	
	speedl=speedr=SLOW_SPEED;
    motorp(LEFT_MOTOR, speedl );
    motorp(RIGHT_MOTOR, SLOW_SPEED+10);
	
	while (FOLLOWING==1)
	{
		LEFT_CDS;
		cds_lt=MUX;
		MIDDLE_CDS;
		cds_md=MUX;
		RIGHT_CDS;
		cds_rt=MUX;	
		
		if (cds_lt<WHITE_LT) adjustLeft();
		if (cds_rt<WHITE_RT) adjustRight();
		
		
		wait(30);
    	
	}
}//end FOLLOW


//ADJUSTLEFT
void adjustLeft(void)
{
	int cds_md,cds_rt,cds_lt;//current cds sensor value
	int c=0; //loop count

  //stop or reverse one motor until the some cds cell
  //  passes over the line, or times out
	motorp(LEFT_MOTOR,speedl=ZERO_SPEED); 
	
	while(c<TIMEOUT)
	{
		MIDDLE_CDS;
		cds_md=MUX;
		RIGHT_CDS;
		cds_rt=MUX;	
		LEFT_CDS;
		cds_lt=MUX;
        
        //if center cds sensor crosses line, go straight again
        if (cds_md<WHITE_MD)
        {
        
        	motorp(LEFT_MOTOR,speedl=SLOW_SPEED );
        	break;
        }	
        
        //if left cds sensor hits, adjust right
        else if (cds_lt<WHITE_LT)
        {
        
            motorp(LEFT_MOTOR,speedl=SLOW_SPEED );
        	adjustRight();
        	break;
        }	
        
        //if right cds cell hit, adjust left
       else if (cds_rt<WHITE_RT)
        {
            motorp(LEFT_MOTOR,speedl=SLOW_SPEED );
        	adjustLeft();
        	break;
        }	        	

		wait(20);
		c++;
	}	
	//if timed out, turn off following flag
	if (c>=TIMEOUT) FOLLOWING=0;
    motorp(LEFT_MOTOR,speedl=SLOW_SPEED );
}//end ADJUSTLEFT


//ADJUSTRIGHT
void adjustRight(void)
{
	int cds_md,cds_rt,cds_lt;//current cds sensor value
	int c=0; //loop count

  //stop or reverse one motor until the some cds cell
  //  passes over the line, or times out
	motorp(RIGHT_MOTOR,speedr=ZERO_SPEED); 
	
	while(c<TIMEOUT)
	{
		MIDDLE_CDS;
		cds_md=MUX;
		RIGHT_CDS;
		cds_rt=MUX;	
		LEFT_CDS;
		cds_lt=MUX;
        
        //if center cds sensor crosses line, go straight again
        if (cds_md<WHITE_MD)
        {
        	motorp(RIGHT_MOTOR,speedr=SLOW_SPEED+10);
        	break;
        	//test lines below
        }	
        
        //if left cds sensor hits, adjust right
        if (cds_lt<WHITE_LT)
        {
          	motorp(RIGHT_MOTOR,speedr=SLOW_SPEED+10);
        	adjustRight();
        	break;
        	//test lines below
        }	
        
        //if right cds cell hit, adjust left
        if (cds_rt<WHITE_RT)
        {
            motorp(RIGHT_MOTOR,speedr=SLOW_SPEED+10);
        	adjustLeft();
        	break;
        	//test lines below
        }	        	

		wait(20);
		c++;
	}	
	//if timed out, turn off following flag
	
    motorp(RIGHT_MOTOR,speedr=SLOW_SPEED+10);
	if (c>=TIMEOUT) FOLLOWING=0;
	
}//end ADJUSTRIGHT


//CALIBRATE
void calibrate(void)
{  
  //calibrate line detection threshhold for cds cells
	int i=0; //loop counter
	WHITE_LT=WHITE_RT=WHITE_MD=0; //initialize sensors values
	
  //for each cds cell: take 3 readings 	
  //with a delay between each.
	while(i<3)
	{
		LEFT_CDS;
		WHITE_LT=WHITE_LT+MUX;
		RIGHT_CDS;
		WHITE_RT=WHITE_RT+MUX;
		MIDDLE_CDS;
		WHITE_MD=WHITE_MD+MUX;
		wait(25);
		i++;
	}
	      
  //Find mean.  
  //Knock off 10 points as a noise buffer,
  //set as threshhold for line detection 
	WHITE_LT=WHITE_LT/3 - 10;
	WHITE_RT=WHITE_RT/3 - 10;
	WHITE_MD=WHITE_MD/3 - 10;	

}//end of CALIBRATE
	

void turn(void)
/************************************************************************
 * Function:  Will turn in a random direction for a "random" amount of  *
 *            time, dictated by the fast changine lower bits in		*
 *	      mseconds().  					        *
 * Returns:  None                                                       *
 *                                                                      *
 * Inputs                                                               *
 *   Parameters: None                                                   *
 *   Globals:    None                                                   *
 *   Registers:  TCNT                                                   *
 * Outputs                                                              *
 *   Parameters: None                                                   *
 *   Globals:    None                                                   *
 *   Registers:  None                                                   *
 * Functions called: motorp(), wait()					*
 * Notes:                                                               *
 ************************************************************************/
{
 int i;
 unsigned rand;

 rand = TCNT;

 if (rand & 0x0001)
 /*turn left*/
 {
   motorp(RIGHT_MOTOR, speedr=MAX_SPEED);
   motorp(LEFT_MOTOR, speedl=-MAX_SPEED);
 }
 else
 /*turn right*/
 {
   motorp(RIGHT_MOTOR, speedr=-MAX_SPEED);
   motorp(LEFT_MOTOR, speedl=MAX_SPEED);
 }

 i=(rand % 1024);
 if(i>250) wait(i); else wait(250);

}

/***********************End Function turn ****************************/


